from vidur.request_generator.gamma_request_interval_generator import (
    GammaRequestIntervalGenerator,
)
from vidur.request_generator.poisson_request_interval_generator import (
    PoissonRequestIntervalGenerator,
)
from vidur.request_generator.static_request_interval_generator import (
    StaticRequestIntervalGenerator,
)
from vidur.request_generator.trace_request_interval_generator import (
    TraceRequestIntervalGenerator,
)
from vidur.types import RequestIntervalGeneratorType
from vidur.utils.base_registry import BaseRegistry


class RequestIntervalGeneratorRegistry(BaseRegistry):
    pass


RequestIntervalGeneratorRegistry.register(
    RequestIntervalGeneratorType.GAMMA, GammaRequestIntervalGenerator
)
RequestIntervalGeneratorRegistry.register(
    RequestIntervalGeneratorType.POISSON, PoissonRequestIntervalGenerator
)
RequestIntervalGeneratorRegistry.register(
    RequestIntervalGeneratorType.STATIC, StaticRequestIntervalGenerator
)
RequestIntervalGeneratorRegistry.register(
    RequestIntervalGeneratorType.TRACE, TraceRequestIntervalGenerator
)



from vidur.entities import Request
from vidur.scheduler.replica_scheduler.mnemosyne_edf_replica_scheduler import (
    MnemosyneEDFReplicaScheduler,
)


class MnemosyneLRSReplicaScheduler(MnemosyneEDFReplicaScheduler):

    def _get_remaining_slack_fraction(self, time: float, request: Request) -> float:
        remaining_prefill_time = self._prefill_time_calculator.get_prefill_time(
            request.num_prefill_tokens,
            self._get_num_processed_tokens(request),
        )
        slack = request.deadline - time - remaining_prefill_time
        return slack / request.deadline_time

    def _sort_request_queue(self, time: float) -> None:
        self._request_queue.sort(
            key=lambda x: self._get_remaining_slack_fraction(time, x[1])
        )



from multiprocessing import Pool
from typing import Callable, Iterable, List, Optional, TypeVar

T = TypeVar("T")
R = TypeVar("R")


def parallel_map(
    func: Callable[[T], R],
    items: Iterable[T],
    num_workers: Optional[int] = None,
    chunk_size: Optional[int] = None,
) -> List[R]:
    """
    Performs parallel mapping of a function over an iterable using multiple processes.

    Args:
        func: The function to apply to each item
        items: Iterable of items to process
        num_workers: Number of worker processes (defaults to number of CPU cores)
        chunk_size: Size of chunks to send to each worker (defaults to len(items)/num_workers)

    Returns:
        List of results in the same order as the input items

    Example:
        >>> def square(x): return x * x
        >>> parallel_map(square, range(10))
        [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]
    """
    with Pool(processes=num_workers) as pool:
        results = pool.map(func, items, chunksize=chunk_size)
    return list(results)



from vidur.entities.batch import Batch
from vidur.entities.batch_stage import BatchStage
from vidur.entities.cluster import Cluster
from vidur.entities.execution_time import ExecutionTime
from vidur.entities.kv_parallel_batch import KVParallelBatch
from vidur.entities.kv_parallel_batch_stage import KVParallelBatchStage
from vidur.entities.replica import Replica
from vidur.entities.request import Request

__all__ = [
    Request,
    Replica,
    Batch,
    Cluster,
    BatchStage,
    ExecutionTime,
    KVParallelBatch,
    KVParallelBatchStage,
]



from typing import Optional

import torch
from sarathi.model_executor.layers.activation import SiluAndMul
from sarathi.model_executor.layers.layernorm import RMSNorm
from sarathi.model_executor.layers.rotary_embedding import get_rope
from sarathi.model_executor.parallel_utils import (
    ColumnParallelLinear,
    RowParallelLinear,
    VocabParallelEmbedding,
)

from vidur.profiling.common.cuda_timer import CudaTimer
from vidur.profiling.common.model_config import ModelConfig

REUSE_MEMORY = True


class CausalSelfAttention(torch.nn.Module):

    def __init__(self, config: ModelConfig, world_size: int):
        super().__init__()
        assert config.embedding_dim % config.num_q_heads == 0
        assert config.embedding_dim % world_size == 0
        assert config.num_q_heads % world_size == 0
        assert config.num_kv_heads % world_size == 0

        self.head_dim = config.embedding_dim // config.num_q_heads
        self.num_q_heads_per_worker = config.num_q_heads // world_size
        self.num_kv_heads_per_worker = config.num_kv_heads // world_size

        self.q_size = self.num_q_heads_per_worker * self.head_dim
        self.kv_size = self.num_kv_heads_per_worker * self.head_dim
        self.scaling = self.head_dim**-0.5

        self.qkv_proj = ColumnParallelLinear(
            config.embedding_dim,
            (config.num_q_heads + 2 * config.num_kv_heads) * self.head_dim,
            bias=config.use_bias or config.use_qkv_bias,
            gather_output=False,
            linear_metric_name="attn_pre_proj",
            world_size=world_size,
        )

        self.o_proj = RowParallelLinear(
            config.num_q_heads * self.head_dim,
            config.embedding_dim,
            bias=config.use_bias,
            input_is_parallel=True,
            reduce_results=False,
            linear_metric_name="attn_post_proj",
            world_size=world_size,
        )
        self.rotary_emb = None
        if isinstance(config.rope_theta, int) or isinstance(config.rope_theta, float):
            self.rotary_emb = get_rope(
                self.head_dim,
                rotary_dim=self.head_dim,
                max_position=config.max_model_len,
                base=config.rope_theta,
                is_neox_style=config.is_neox_style,
                rope_scaling=config.rope_scaling,
            )
        self._attn_rope_timer = CudaTimer("attn_rope")

    def forward(self, hidden_states, positions):
        qkv, _ = self.qkv_proj(hidden_states)
        q, k, v = qkv.split([self.q_size, self.kv_size, self.kv_size], dim=-1)
        with self._attn_rope_timer:
            q, k = self.rotary_emb(positions, q, k)
        # output from attn has the same shape as q
        attn_output = torch.randn_like(q)
        output, _ = self.o_proj(attn_output)
        return output


class MLP(torch.nn.Module):
    def __init__(self, config: ModelConfig, world_size: int):
        super().__init__()

        assert config.embedding_dim % world_size == 0

        if config.use_gated_mlp:
            self.up_proj = ColumnParallelLinear(
                config.embedding_dim,
                2 * config.mlp_hidden_dim,
                bias=config.use_bias,
                gather_output=False,
                world_size=world_size,
                linear_metric_name="mlp_up_proj",
            )
            self.act = SiluAndMul()
        else:
            self.up_proj = ColumnParallelLinear(
                config.embedding_dim,
                config.mlp_hidden_dim,
                bias=config.use_bias,
                gather_output=False,
                world_size=world_size,
                linear_metric_name="mlp_up_proj",
            )
            self.act = torch.nn.GELU()

        self.down_proj = RowParallelLinear(
            config.mlp_hidden_dim,
            config.embedding_dim,
            bias=config.use_bias,
            input_is_parallel=True,
            world_size=world_size,
            reduce_results=False,
            linear_metric_name="mlp_down_proj",
        )

        self.mlp_act_timer = CudaTimer("mlp_act")

    def forward(self, hidden_states):
        hidden_states, _ = self.up_proj(hidden_states)
        with self.mlp_act_timer:
            hidden_states = self.act(hidden_states)
        hidden_states, _ = self.down_proj(hidden_states)
        return hidden_states


class GPTBlock(torch.nn.Module):

    def __init__(self, config: ModelConfig, world_size: int):
        super().__init__()

        if config.norm == "layer_norm":
            self.input_layernorm = torch.nn.LayerNorm(config.embedding_dim)
        elif config.norm == "rms_norm":
            self.input_layernorm = RMSNorm(config.embedding_dim)
        else:
            raise ValueError(f"Unknown norm: {config.norm} for input_layernorm")

        self._post_attn_norm = config.post_attn_norm
        if config.post_attn_norm:
            if config.norm == "rms_norm":
                self.post_attention_layernorm = RMSNorm(config.embedding_dim)
            else:
                raise ValueError(
                    f"Unknown norm: {config.norm} for post_attention_layernorm"
                )

        self.attn = CausalSelfAttention(config, world_size)
        self.mlp = MLP(config, world_size)

        self.input_layernorm_timer = CudaTimer("input_layernorm")
        self.post_attention_layernorm_timer = CudaTimer("post_attention_layernorm")
        self.add_timer = CudaTimer("add")

    def forward(self, positions, hidden_states, residual):
        if self._post_attn_norm:
            return self._forward_with_post_attn_norm(positions, hidden_states, residual)
        else:
            return self._forward_without_post_attn_norm(positions, hidden_states)

    def _forward_with_post_attn_norm(
        self,
        positions: torch.Tensor,
        hidden_states: torch.Tensor,
        residual: Optional[torch.Tensor],
    ):
        # Self Attention
        residual = hidden_states
        with self.input_layernorm_timer:
            hidden_states = self.input_layernorm(hidden_states)
        hidden_states = self.attn(
            positions=positions,
            hidden_states=hidden_states,
        )
        hidden_states = residual + hidden_states
        # Fully Connected
        residual = hidden_states
        with self.post_attention_layernorm_timer:
            hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        with self.add_timer:
            hidden_states = residual + hidden_states
        return hidden_states

    def _forward_without_post_attn_norm(
        self,
        positions: torch.Tensor,
        hidden_states: torch.Tensor,
    ):
        residual = hidden_states
        with self.input_layernorm_timer:
            hidden_states = self.input_layernorm(hidden_states)
        attn_outputs = self.attn(
            positions=positions,
            hidden_states=hidden_states,
        )
        feed_forward_hidden_states = self.mlp(hidden_states)
        with self.add_timer:
            hidden_states = attn_outputs + feed_forward_hidden_states + residual
        return hidden_states


class GPTModel(torch.nn.Module):
    def __init__(self, config: ModelConfig, world_size: int, num_repeat_steps: int = 1):
        super().__init__()

        self.num_repeat_steps = num_repeat_steps

        self.embed_tokens = VocabParallelEmbedding(
            config.vocab_size,
            config.embedding_dim,
            linear_metric_name="emb",
            reduce_results=False,
            world_size=world_size,
            rank=0,
        )

        self.block = GPTBlock(config, world_size=world_size)

    def forward(self, input_ids, positions):
        hidden_states = self.embed_tokens(input_ids)
        residual = hidden_states
        for _ in range(self.num_repeat_steps):
            hidden_states = self.embed_tokens(input_ids)
            hidden_states = self.block(
                positions,
                hidden_states,
                residual,
            )

        return hidden_states



from typing import List

from vidur.events import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class ReplicaScheduleEvent(BaseEvent):
    def __init__(self, time: float, replica_id: int):
        super().__init__(time, EventType.REPLICA_SCHEDULE)

        self._replica_id = replica_id

        self._batches = []

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.batch_stage_arrival_event import BatchStageArrivalEvent

        replica_scheduler = scheduler.get_replica_scheduler(self._replica_id)
        self._batch = replica_scheduler.on_schedule(self._time)

        if not self._batch:
            return []

        memory_usage_percent = replica_scheduler.memory_usage_percent
        metrics_store.on_replica_schedule(
            self.time, self._replica_id, memory_usage_percent
        )

        self._batch.on_schedule(self.time)

        return [
            BatchStageArrivalEvent(
                self.time,
                self._replica_id,
                0,  # stage_id
                self._batch,
            )
        ]

    def to_dict(self):
        return {
            "time": self.time,
            "event_type": self.event_type,
            "replica_id": self._replica_id,
        }



import argparse
import glob
import json
import os
from multiprocessing import Pool

import numpy as np
import pandas as pd
import yaml

from vidur.config_optimizer.analyzer.constants import CPU_MACHINE_COST, GPU_COSTS
from vidur.logger import init_logger

logger = init_logger(__name__)


def extract_stat_from_request_metrics(
    request_metrics_df: pd.DataFrame,
    stat_name: str,
    stat_short_name: str = None,
):
    if stat_short_name is None:
        stat_short_name = stat_name

    stats = request_metrics_df[stat_name].describe().to_dict()
    # add 95th and 99th percentile
    stats["90%"] = request_metrics_df[stat_name].quantile(0.90)
    stats["95%"] = request_metrics_df[stat_name].quantile(0.95)
    stats["99%"] = request_metrics_df[stat_name].quantile(0.99)

    stats_dict = {f"{stat_short_name}_{k}": v for k, v in stats.items()}
    return stats_dict


def extract_stats_from_cdf_df(
    cdf_df: pd.DataFrame,
    stat_name: str,
    stat_short_name: str = None,
    extract_all: bool = False,
):
    if stat_short_name is None:
        stat_short_name = stat_name

    if extract_all:
        cdf_df["cdf_rounded"] = cdf_df["cdf"].round(2)
        cdf_df = cdf_df.drop_duplicates(subset="cdf_rounded", keep="first")
        return {f"{stat_short_name}_cdf": cdf_df[stat_name].tolist()[1:]}

    percentile_map = {
        "min": 0.0,
        "25%": 0.25,
        "50%": 0.5,
        "75%": 0.75,
        "90%": 0.90,
        "95%": 0.95,
        "99%": 0.99,
        "max": 1.0,
    }
    stats = {
        k: cdf_df[cdf_df["cdf"] == v][stat_name].iloc[0]
        for k, v in percentile_map.items()
    }
    stats_dict = {f"{stat_short_name}_{k}": v for k, v in stats.items()}
    return stats_dict


def extract_utilization_stats(run_dir: str, stat_name: str):
    stat_files = glob.glob(f"{run_dir}/plots/replica_*{stat_name}.json")
    vals = []
    for stat_file in stat_files:
        stat = json.load(open(stat_file))
        for k, v in stat.items():
            if k.endswith("weighted_mean"):
                vals.append(v)

    if len(vals) == 0:
        return {f"{stat_name}_mean": np.nan}

    return {f"{stat_name}_mean": sum(vals) / len(vals)}


def process_run(run_dir: str):
    config_file = f"{run_dir}/config.yml"
    request_metrics_file = f"{run_dir}/request_metrics.csv"
    tbt_file = f"{run_dir}/plots/batch_execution_time.csv"
    ttft_file = f"{run_dir}/plots/prefill_e2e_time.csv"
    batch_size_file = f"{run_dir}/plots/batch_size.csv"
    batch_num_tokens_file = f"{run_dir}/plots/batch_num_tokens.csv"
    request_completion_time_series_file = (
        f"{run_dir}/plots/request_completion_time_series.csv"
    )

    try:
        with open(config_file, "r") as f:
            config = yaml.safe_load(f)

        request_metrics_df = pd.read_csv(request_metrics_file)
        tbt_df = pd.read_csv(tbt_file)
        ttft_df = pd.read_csv(ttft_file)
        batch_size_df = pd.read_csv(batch_size_file)
        batch_num_tokens_df = pd.read_csv(batch_num_tokens_file)
        request_completion_time_series_df = pd.read_csv(
            request_completion_time_series_file
        )
    except FileNotFoundError as e:
        # TODO(amey): Add a better error handling approach
        # we can run into this issue if the run was not successful
        # either due to actual failure or due to model OOMing in simulation
        # later is okay, while the former is not
        return None

    request_scheduling_delay_stats = extract_stat_from_request_metrics(
        request_metrics_df, "request_scheduling_delay"
    )
    request_e2e_time_normalized_stats = extract_stat_from_request_metrics(
        request_metrics_df, "request_e2e_time_normalized"
    )
    ttft_stats = extract_stat_from_request_metrics(
        request_metrics_df, "prefill_e2e_time", "ttft"
    )
    ttft_cdf = extract_stats_from_cdf_df(
        ttft_df, "prefill_e2e_time", "ttft", extract_all=True
    )
    tbt_stats = extract_stats_from_cdf_df(tbt_df, "batch_execution_time", "tbt")
    tbt_cdf = extract_stats_from_cdf_df(
        tbt_df, "batch_execution_time", "tbt", extract_all=True
    )
    batch_size_stats = extract_stats_from_cdf_df(batch_size_df, "batch_size")
    batch_size_cdf = extract_stats_from_cdf_df(
        batch_size_df, "batch_size", extract_all=True
    )
    batch_num_tokens_cdf = extract_stats_from_cdf_df(
        batch_num_tokens_df, "batch_num_tokens", extract_all=True
    )
    memory_usage_stats = extract_utilization_stats(run_dir, "memory_usage")
    mfu_stats = extract_utilization_stats(run_dir, "mfu")
    busy_time_percent_stats = extract_utilization_stats(run_dir, "busy_time_percent")
    runtime = request_completion_time_series_df["Time (sec)"].max()

    if (
        config["replica_scheduler_provider"] == "sarathi"
        and config["sarathi_scheduler_chunk_size"] == 4096
    ):
        config["replica_scheduler_provider"] = "orca+"

    config.update(
        {
            **request_scheduling_delay_stats,
            **request_e2e_time_normalized_stats,
            **tbt_stats,
            **ttft_stats,
            **memory_usage_stats,
            **mfu_stats,
            **busy_time_percent_stats,
            **ttft_cdf,
            **tbt_cdf,
            **batch_size_stats,
            **batch_size_cdf,
            **batch_num_tokens_cdf,
            "runtime": runtime,
        }
    )

    return config


def get_sim_time(sim_results_dir: str):
    output_file = f"{sim_results_dir}/output.log"

    with open(output_file, "r") as f:
        lines = f.readlines()

    # search for Simulation took time: xxx
    for line in lines:
        if "Simulation took time" in line:
            return float(line.split(":")[-1].strip())


def process_trace(sim_results_dir: str):
    analysis_dir = f"{sim_results_dir}/analysis"

    # check the results already exists
    if os.path.exists(f"{analysis_dir}/stats.csv") and os.path.exists(
        f"{analysis_dir}/simulation_stats.yml"
    ):
        return

    os.makedirs(analysis_dir, exist_ok=True)

    # the dir structure is sim_results_dir/runs/<config_hash>/<qps>/<date-string>/
    run_dirs = glob.glob(f"{sim_results_dir}/runs/*/*/*/")

    num_cores = os.cpu_count() - 2

    with Pool(num_cores) as p:
        all_results = p.map(process_run, run_dirs)

    # filer out None values
    all_results = [r for r in all_results if r is not None]
    logger.info(f"Total number of runs: {len(run_dirs)} valid runs: {len(all_results)}")

    df = pd.DataFrame(all_results)

    df["num_gpus"] = (
        df["cluster_num_replicas"]
        * df["replica_num_tensor_parallel_workers"]
        * df["replica_num_pipeline_stages"]
    )
    df["cost"] = (
        df["runtime"] * df["num_gpus"] * df["replica_device"].map(GPU_COSTS) / 3600
    )
    df["capacity_per_dollar"] = df["poisson_request_interval_generator_qps"] / (
        df["num_gpus"] * df["replica_device"].map(GPU_COSTS)
    )
    df["gpu_hrs"] = df["runtime"] * df["num_gpus"] / 3600

    df["num_replica_gpus"] = (
        df["replica_num_tensor_parallel_workers"] * df["replica_num_pipeline_stages"]
    )
    df["hour_cost_per_replica"] = (
        df["replica_device"].map(GPU_COSTS) * df["num_replica_gpus"]
    )
    df["capacity_per_replica"] = (
        df["poisson_request_interval_generator_qps"] / df["cluster_num_replicas"]
    )

    # store the df
    df.to_csv(f"{analysis_dir}/stats.csv", index=False)

    gpu_cost = df["cost"].sum()
    total_gpu_hrs = df["gpu_hrs"].sum()

    sim_time = get_sim_time(sim_results_dir)
    cpu_hrs = sim_time / 3600
    cpu_cost = cpu_hrs * CPU_MACHINE_COST

    simulation_stats = {
        "gpu_cost": gpu_cost,
        "sim_cpu_cost": cpu_cost,
        "total_gpu_hrs": total_gpu_hrs,
        "sim_time": sim_time,
        "total_runs": len(run_dirs),
        "valid_runs": len(all_results),
    }

    json.dump(
        simulation_stats, open(f"{analysis_dir}/simulation_stats.json", "w"), indent=4
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--sim-results-dir", type=str, required=True)
    args = parser.parse_args()

    process_trace(args.sim_results_dir)


if __name__ == "__main__":
    main()



import math
import random
from typing import Tuple

from vidur.request_generator.base_request_length_generator import (
    BaseRequestLengthGenerator,
)


class UniformRequestLengthGenerator(BaseRequestLengthGenerator):

    def get_next_num_tokens(self) -> Tuple[float, float]:
        total_tokens = random.uniform(
            self.config.min_tokens,
            self.config.max_tokens,
        )

        decode_tokens = math.ceil(
            total_tokens / (1 + self.config.prefill_to_decode_ratio)
        )
        prefill_tokens = total_tokens - decode_tokens
        assert prefill_tokens > 0 and decode_tokens > 0

        return prefill_tokens, decode_tokens



import argparse
import os

import matplotlib.colors as mcolors
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.colors import LinearSegmentedColormap
from paretoset import paretoset

from vidur.config_optimizer.analyzer.constants import (
    AXIS_COLS,
    CONFIG_KEY,
    METRICS,
    PRETTY_NAMES,
)
from vidur.config_optimizer.analyzer.utils import get_trace_name


def plot_pareto_curve_under_slos(
    df,
    model_name: str,
    trace_file: str,
    results_dir: str,
    metric_1: str,
    percentile_1: str,
    slo_1: float,
    metric_2: str,
    percentile_2: str,
    slo_2: float,
):
    metric_1_col = f"{metric_1}_{percentile_1}%"
    metric_2_col = f"{metric_2}_{percentile_2}%"

    # filter out the rows that are not in the pareto frontier
    metric_1_paretoset_mask = paretoset(
        df[["capacity_per_dollar", metric_1_col]], sense=["max", "min"]
    )
    metric_2_paretoset_mask = paretoset(
        df[["capacity_per_dollar", metric_2_col]], sense=["max", "min"]
    )

    metric_1_pareto_df = df[metric_1_paretoset_mask]
    metric_2_pareto_df = df[metric_2_paretoset_mask]

    # filter based on the SLOs
    slo_mask = (df[metric_1_col] <= slo_1) & (df[metric_2_col] <= slo_2)
    slo_df = df[slo_mask]

    fig, axs = plt.subplots(1, 3, figsize=(20, 5))

    # Legends for plots
    red_line = mlines.Line2D([], [], color="red", label="SLO Limit")
    green_patch = mpatches.Patch(color="green", alpha=0.8, label="SLO Compliant Region")
    orange_scatter = mlines.Line2D(
        [],
        [],
        color="orange",
        marker="o",
        linestyle="None",
        markersize=10,
        alpha=0.95,
        label="Near Pareto Configs",
    )
    green_scatter = mlines.Line2D(
        [],
        [],
        color="green",
        marker="o",
        linestyle="None",
        markersize=10,
        alpha=0.95,
        label="SLO Compliant Configs",
    )
    green_star = mlines.Line2D(
        [],
        [],
        color="green",
        marker="*",
        linestyle="None",
        markersize=15,
        alpha=0.95,
        label="Best Config",
    )

    # plot the pareto curve for metric_1
    axs[0].scatter(df[metric_1_col], df["capacity_per_dollar"], color="blue", alpha=0.8)
    # draw the SLO line
    axs[0].axvline(x=slo_1, color="red", alpha=0.7)
    # add light green background for the points that satisfy the SLO
    axs[0].axvspan(0, slo_1, alpha=0.1, color="green")
    # draw pareto curve
    metric_1_pareto_df = metric_1_pareto_df.sort_values(metric_1_col)
    axs[0].plot(
        metric_1_pareto_df[metric_1_col],
        metric_1_pareto_df["capacity_per_dollar"],
        color="orange",
    )
    # draw the points that satisfy the SLO
    axs[0].scatter(
        slo_df[metric_1_col], slo_df["capacity_per_dollar"], color="green", alpha=0.8
    )

    # plot the pareto curve for metric_2
    axs[1].scatter(df[metric_2_col], df["capacity_per_dollar"], color="blue", alpha=0.8)
    # draw the SLO line
    axs[1].axvline(x=slo_2, color="red", alpha=0.7)
    # add light green background for the points that satisfy the SLO
    axs[1].axvspan(0, slo_2, alpha=0.1, color="green")
    # draw pareto curve
    metric_2_pareto_df = metric_2_pareto_df.sort_values(metric_2_col)
    axs[1].plot(
        metric_2_pareto_df[metric_2_col],
        metric_2_pareto_df["capacity_per_dollar"],
        color="orange",
    )
    # draw the points that satisfy the SLO
    axs[1].scatter(
        slo_df[metric_2_col], slo_df["capacity_per_dollar"], color="green", alpha=0.8
    )

    # plot the pareto curve for metric_1 and metric_2
    # plot all the points
    norm = mcolors.Normalize(
        vmin=slo_df["capacity_per_dollar"].min(),
        vmax=slo_df["capacity_per_dollar"].max(),
    )
    blue_to_orange_cmap = LinearSegmentedColormap.from_list("", ["blue", "orange"])
    axs[2].scatter(
        df[metric_1_col],
        df[metric_2_col],
        alpha=0.95,
        c=df["capacity_per_dollar"],
        norm=norm,
        cmap=blue_to_orange_cmap,
    )

    # add colorbar
    cbar = plt.colorbar(axs[2].collections[0], ax=axs[2])
    cbar.set_label("QPS per Dollar", rotation=270, labelpad=20, fontsize=14)

    # plot the SLOs
    axs[2].axvline(x=slo_1, color="red", alpha=0.7)
    axs[2].axhline(y=slo_2, color="red", alpha=0.7)

    # axs[2].set_xlim(0, min(1.5 * slo_1, metric_1_pareto_df[metric_1_col].max() * 1.25))
    # axs[2].set_ylim(0, min(1.5 * slo_2, metric_2_pareto_df[metric_2_col].max() * 1.25))
    axs[2].set_xlim(
        max(0.25 * slo_1, metric_1_pareto_df[metric_1_col].min() * 1.2),
        min(2 * slo_1, metric_1_pareto_df[metric_1_col].max() * 1.5),
    )
    axs[2].set_ylim(0, slo_2 * 1.5)

    # add light green background for the region that satisfies both the SLOs
    axs[2].axvspan(
        0, slo_1, alpha=0.15, color="green", ymin=0, ymax=slo_2 / axs[2].get_ylim()[1]
    )

    # highlight the best point i.e. max capacity per dollar under the SLOs
    best_points = slo_df.sort_values("capacity_per_dollar", ascending=False)
    if best_points.shape[0] > 0:
        best_point = best_points.iloc[0]
        # get the star color based on the capacity per dollar
        star_color = blue_to_orange_cmap(norm(best_point["capacity_per_dollar"]))
        # update the color in legend
        green_star.set_color(star_color)
        axs[2].scatter(
            best_point[metric_1_col],
            best_point[metric_2_col],
            color=star_color,
            s=800,
            marker="*",
            alpha=0.95,
        )

    axs[0].legend(
        handles=[red_line, green_patch, green_scatter, green_star],
        # loc="upper left",
        loc="upper right",
        fontsize=14,
    )

    # set the x and y labels
    axs[0].set_xlabel(f"{METRICS[metric_1]} - P{percentile_1} (s)")
    axs[0].set_ylabel("QPS per Dollar")
    axs[1].set_xlabel(f"{METRICS[metric_2]} - P{percentile_2} (s)")
    axs[1].set_ylabel("QPS per Dollar")
    axs[2].set_xlabel(f"{METRICS[metric_1]} - P{percentile_1} (s)")
    axs[2].set_ylabel(f"{METRICS[metric_2]} - P{percentile_2} (s)")
    # increase the font size of the labels
    for ax in axs:
        ax.xaxis.label.set_size(18)
        ax.yaxis.label.set_size(18)
        # increase tick font size
        ax.tick_params(axis="both", which="major", labelsize=14)

    # set the x and y limits
    axs[0].set_xlim(
        max(0.25 * slo_1, metric_1_pareto_df[metric_1_col].min() * 1.2),
        min(2 * slo_1, metric_1_pareto_df[metric_1_col].max() * 1.5),
    )
    axs[1].set_xlim(0.1, slo_2 * 1.25)

    # write the best config axis values at the bottom
    if best_points.shape[0] > 0:
        best_point_str = ""
        for axis_col, axis_name in AXIS_COLS.items():
            if (
                axis_col == "sarathi_scheduler_chunk_size"
                and best_point["replica_scheduler_provider"] != "sarathi"
            ):
                continue
            value = best_point[axis_col]
            value = str(value) if not value in PRETTY_NAMES else PRETTY_NAMES[value]
            best_point_str += f"{axis_name}: {value}, "

        best_point_str = best_point_str[:-2]
        best_point_str += f"\nQPS per Dollar: {best_point['capacity_per_dollar']:.2f}"
        fig.suptitle(f"Best Config: {best_point_str}", fontsize=20)

    # tight plot layout
    plt.tight_layout()

    # save the plot
    output_dir = f"{results_dir}/pareto_curves/{model_name}"
    os.makedirs(output_dir, exist_ok=True)
    plt.savefig(
        f"{output_dir}/{trace_file}_{metric_1}_{metric_2}_{percentile_1}_{percentile_2}_{slo_1}_{slo_2}.png"
    )
    plt.savefig(
        f"{output_dir}/{trace_file}_{metric_1}_{metric_2}_{percentile_1}_{percentile_2}_{slo_1}_{slo_2}.pdf"
    )


def select_best_run(
    df: pd.DataFrame,
    scheduling_delay_slo_percentile: float,
    scheduling_delay_slo_value: float,
):
    metric_names = f"request_scheduling_delay_{scheduling_delay_slo_percentile}%"
    mask = df[metric_names] < scheduling_delay_slo_value
    if len(df[mask]) > 0:
        # multiple points satisfy the SLO
        # select the point with the highest qps
        return (
            df[mask]
            .sort_values("poisson_request_interval_generator_qps", ascending=False)
            .iloc[0]
        )
    else:
        # no point satisfy the SLO
        # select the point with the lowest scheduling delay
        return df.iloc[df[metric_names].argmin()]


def process_sim_results(args: argparse.Namespace):
    analysis_dir = f"{args.sim_results_dir}/analysis"

    stats_file = f"{analysis_dir}/stats.csv"

    if not os.path.exists(stats_file):
        raise ValueError(f"Stats file {stats_file} does not exist")

    df = pd.read_csv(stats_file)
    df = df[df["sarathi_scheduler_chunk_size"] != 256]
    df = df[df["replica_device"] != "a40"]

    OLD_GPU_COSTS = {
        "h100": 4.25,
        "a100": 2.21,
        "a40": 1.28,
    }

    NEW_GPU_COSTS = {
        "h100": 6.98,
        "a100": 3.72,
        "a40": 1.28,
    }

    # update the "cost" and "capacity_per_dollar" columns accourding to the new costs use "replica_device" as the key
    def update_cost(row):
        if row["replica_device"] == "a40":
            return row["cost"]
        return (
            NEW_GPU_COSTS[row["replica_device"]]
            * row["cost"]
            / OLD_GPU_COSTS[row["replica_device"]]
        )

    def update_capacity_per_dollar(row):
        if row["replica_device"] == "a40":
            return row["capacity_per_dollar"]
        return (
            row["capacity_per_dollar"]
            * OLD_GPU_COSTS[row["replica_device"]]
            / NEW_GPU_COSTS[row["replica_device"]]
        )

    df["cost"] = df.apply(update_cost, axis=1)
    df["capacity_per_dollar"] = df.apply(update_capacity_per_dollar, axis=1)

    for (model_name, trace_file), group in df.groupby(
        ["replica_model_name", "trace_request_length_generator_trace_file"]
    ):
        subset_df = (
            group.groupby(CONFIG_KEY)
            .apply(
                lambda x: select_best_run(
                    x,
                    args.scheduling_delay_slo_percentile,
                    args.scheduling_delay_slo_value,
                )
            )
            .reset_index(drop=True)
        )

        trace_name = get_trace_name(trace_file)

        plot_pareto_curve_under_slos(
            subset_df,
            model_name,
            trace_name,
            analysis_dir,
            "ttft",
            args.ttft_slo_percentile,
            args.ttft_slo_value,
            "tbt",
            args.tbt_slo_percentile,
            args.tbt_slo_value,
        )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--sim-results-dir", type=str, required=True)
    parser.add_argument("--scheduling-delay-slo-percentile", type=float, default=99)
    parser.add_argument("--scheduling-delay-slo-value", type=float, default=5.0)
    parser.add_argument("--ttft-slo-percentile", type=float, default=90)
    parser.add_argument("--ttft-slo-value", type=float, default=2.0)
    parser.add_argument("--tbt-slo-percentile", type=float, default=99)
    parser.add_argument("--tbt-slo-value", type=float, default=0.2)
    args = parser.parse_args()

    process_sim_results(args)


if __name__ == "__main__":
    main()



from typing import List, Union

from vidur.entities import Batch, BatchStage, KVParallelBatch, KVParallelBatchStage
from vidur.events import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class BatchStageEndEvent(BaseEvent):
    def __init__(
        self,
        time: float,
        replica_id: int,
        stage_id: int,
        is_last_stage: bool,
        batch: Union[Batch, KVParallelBatch],
        batch_stage: Union[BatchStage, KVParallelBatchStage],
    ):
        super().__init__(time, EventType.BATCH_STAGE_END)

        self._replica_id = replica_id
        self._stage_id = stage_id
        self._is_last_stage = is_last_stage

        self._batch = batch
        self._batch_stage = batch_stage

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.batch_end_event import BatchEndEvent
        from vidur.events.batch_stage_arrival_event import BatchStageArrivalEvent
        from vidur.events.replica_schedule_event import ReplicaScheduleEvent
        from vidur.events.replica_stage_schedule_event import ReplicaStageScheduleEvent

        self._batch_stage.on_stage_end(self.time)
        scheduler.get_replica_stage_scheduler(
            self._replica_id, self._stage_id
        ).on_stage_end()
        scheduler.get_replica_scheduler(self._replica_id).on_stage_end(
            self._stage_id, self._batch_stage
        )
        metrics_store.on_batch_stage_end(
            self.time,
            self._batch_stage,
            self._replica_id,
            self._stage_id,
        )

        next_events = [
            ReplicaStageScheduleEvent(
                self.time,
                self._replica_id,
                self._stage_id,
            ),
        ]

        if self._is_last_stage:
            next_events += [BatchEndEvent(self.time, self._replica_id, self._batch)]
        else:
            # schedule the next stage for the current batch if sequence pipeline parallel is not enabled
            next_events += [
                BatchStageArrivalEvent(
                    self.time,
                    self._replica_id,
                    self._stage_id + 1,
                    self._batch,
                )
            ]

        if self._stage_id == 0:
            next_events += [ReplicaScheduleEvent(self.time, self._replica_id)]

        return next_events

    def to_dict(self):
        return {
            "time": self.time,
            "event_type": self.event_type,
            "replica_id": self._replica_id,
            "stage_id": self._stage_id,
            "batch_id": self._batch.id,
            "batch_stage_id": self._batch_stage.id,
            "is_last_stage": self._is_last_stage,
        }

    def to_chrome_trace_events(self) -> dict:
        return self._batch_stage.to_chrome_trace_events(self.time)



from typing import List

from vidur.events import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class GlobalScheduleEvent(BaseEvent):
    def __init__(self, time: float):
        super().__init__(time, EventType.GLOBAL_SCHEDULE)

        self._replica_set = []
        self._request_mapping = []

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.replica_schedule_event import ReplicaScheduleEvent

        self._replica_set = set()
        self._request_mapping = scheduler.schedule()

        for replica_id, request in self._request_mapping:
            self._replica_set.add(replica_id)
            scheduler.get_replica_scheduler(replica_id).add_request(request)

        return [
            ReplicaScheduleEvent(self.time, replica_id)
            for replica_id in self._replica_set
        ]

    def to_dict(self):
        return {
            "time": self.time,
            "event_type": self.event_type,
            "replica_set": self._replica_set,
            "request_mapping": [
                (replica_id, request.id)
                for replica_id, request in self._request_mapping
            ],
        }



from typing import Tuple, Union

from vidur._native.entities import ExecutionTime as ExecutionTimeC
from vidur.entities import Batch, BatchStage, KVParallelBatch
from vidur.execution_time_predictor import BaseExecutionTimePredictor


class ReplicaStageScheduler:
    def __init__(
        self,
        replica_id: int,
        stage_id: int,
        is_last_stage: bool,
        execution_time_predictor: BaseExecutionTimePredictor,
    ) -> None:
        self._replica_id = replica_id
        self._stage_id = stage_id
        self._is_last_stage = is_last_stage
        self._execution_time_predictor = execution_time_predictor

        self._batch_queue = []
        self._is_busy = False

    @property
    def is_last_stage(self) -> bool:
        return self._is_last_stage

    def is_empty(self) -> bool:
        return len(self._batch_queue) == 0

    def add_batch(self, batch: Union[Batch, KVParallelBatch]) -> None:
        self._batch_queue.append(batch)

    def on_stage_end(self) -> None:
        self._is_busy = False

    def on_schedule(
        self,
    ) -> Tuple[Union[Batch, KVParallelBatch], BatchStage, ExecutionTimeC]:
        if self._is_busy or not self._batch_queue:
            return None, None, None

        self._is_busy = True
        batch = self._batch_queue.pop(0)
        execution_time = self._execution_time_predictor.get_execution_time(
            batch,
            self._stage_id,
        )
        batch_stage = batch.to_batch_stage(self._stage_id, execution_time)

        return batch, batch_stage, execution_time



from dataclasses import fields, is_dataclass
from typing import Union, get_args, get_origin

primitive_types = {int, str, float, bool, type(None)}


def get_all_subclasses(cls):
    subclasses = cls.__subclasses__()
    return subclasses + [g for s in subclasses for g in get_all_subclasses(s)]


def is_primitive_type(field_type: type) -> bool:
    # Check if the type is a primitive type
    return field_type in primitive_types


def is_generic_composed_of_primitives(field_type: type) -> bool:
    origin = get_origin(field_type)
    if origin in {list, dict, tuple, Union}:
        # Check all arguments of the generic type
        args = get_args(field_type)
        return all(is_composed_of_primitives(arg) for arg in args)
    return False


def is_composed_of_primitives(field_type: type) -> bool:
    # Check if the type is a primitive type
    if is_primitive_type(field_type):
        return True

    # Check if the type is a generic type composed of primitives
    if is_generic_composed_of_primitives(field_type):
        return True

    return False


def to_snake_case(name: str) -> str:
    return "".join(["_" + i.lower() if i.isupper() else i for i in name]).lstrip("_")


def is_optional(field_type: type) -> bool:
    return get_origin(field_type) is Union and type(None) in get_args(field_type)


def is_list(field_type: type) -> bool:
    # Check if the field type is a List
    return get_origin(field_type) is list


def is_dict(field_type: type) -> bool:
    # Check if the field type is a Dict
    return get_origin(field_type) is dict


def is_bool(field_type: type) -> bool:
    return field_type is bool


def get_inner_type(field_type: type) -> type:
    return next(t for t in get_args(field_type) if t is not type(None))


def is_subclass(cls, parent: type) -> bool:
    return hasattr(cls, "__bases__") and parent in cls.__bases__


def dataclass_to_dict(obj):
    if isinstance(obj, list):
        return [dataclass_to_dict(item) for item in obj]
    elif is_dataclass(obj):
        data = {}
        for field in fields(obj):
            value = getattr(obj, field.name)
            data[field.name] = dataclass_to_dict(value)
        # Include members created in __post_init__
        for key, value in obj.__dict__.items():
            if key not in data:
                data[key] = dataclass_to_dict(value)
        # Include the name of the class
        if hasattr(obj, "get_type") and callable(getattr(obj, "get_type")):
            data["name"] = str(obj.get_type())
        elif hasattr(obj, "get_name") and callable(getattr(obj, "get_name")):
            data["name"] = obj.get_name()
        return data
    else:
        return obj



import json
import shutil
import tempfile
from datetime import datetime
from pathlib import Path

from huggingface_hub import HfApi, upload_folder

from vidur.utils import sanitize_name


def upload_dataset(
    repo_id: str,
    config: dict,
    data_path: str,
    use_private_repo: bool = True,
    commit_message: str = "Published dataset",
):
    api = HfApi()

    # Create repo with Vidur naming conventions
    api.create_repo(
        repo_id=repo_id, repo_type="dataset", exist_ok=True, private=use_private_repo
    )

    # Create temp directory with data + config
    with tempfile.TemporaryDirectory() as tmp_dir:
        # Copy user's data
        shutil.copytree(data_path, tmp_dir, dirs_exist_ok=True)

        # Add config.json
        config_path = Path(tmp_dir) / "config.json"
        with open(config_path, "w") as f:
            json.dump(config, f, indent=2)

        upload_folder(
            folder_path=tmp_dir,
            repo_id=repo_id,
            repo_type="dataset",
            commit_message=commit_message,
        )



from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from vidur.config import SimulationConfig
from vidur.entities import Replica, Request
from vidur.execution_time_predictor import ExecutionTimePredictorRegistry
from vidur.logger import init_logger
from vidur.scheduler.replica_scheduler.replica_scheduler_registry import (
    ReplicaSchedulerRegistry,
)
from vidur.types import ReplicaSchedulerType

logger = init_logger(__name__)


class BaseGlobalScheduler(ABC):
    def __init__(self, config: SimulationConfig, replicas: Dict[int, Replica]):
        self._config = config
        self._replicas = replicas

        self._num_replicas = len(self._replicas)

        logger.info(f"Number of replicas: {self._num_replicas}")

        execution_time_predictor = ExecutionTimePredictorRegistry.get(
            config.execution_time_predictor_config.get_type(),
            predictor_config=config.execution_time_predictor_config,
            replica_config=config.cluster_config.replica_config,
        )
        self._replica_schedulers = {
            replica_id: ReplicaSchedulerRegistry.get(
                config.cluster_config.replica_scheduler_config.get_type(),
                replica_config=config.cluster_config.replica_config,
                replica_scheduler_config=config.cluster_config.replica_scheduler_config,
                request_generator_config=config.request_generator_config,
                replica=replica,
                num_stages=replica.num_pipeline_stages,
                execution_time_predictor=execution_time_predictor,
            )
            for replica_id, replica in replicas.items()
        }
        self._request_queue = []

    def sort_requests(self) -> None:
        self._request_queue.sort(key=lambda request: request._arrived_at)

    def add_request(self, request: Request) -> None:
        self._request_queue.append(request)

    def get_replica_scheduler(self, replica_id: int):
        return self._replica_schedulers[replica_id]

    def get_replica_stage_scheduler(self, replica_id: int, stage_id: int):
        return self._replica_schedulers[replica_id].get_replica_stage_scheduler(
            stage_id
        )

    def is_empty(self) -> bool:
        return len(self._request_queue) == 0 and all(
            replica_scheduler.is_empty()
            for replica_scheduler in self._replica_schedulers.values()
        )

    def get_pending_requests(self) -> List[Request]:
        return self._request_queue + sum(
            [
                replica_scheduler.get_pending_requests()
                for replica_scheduler in self._replica_schedulers.values()
            ],
            [],
        )

    @abstractmethod
    def schedule(self) -> List[Tuple[int, Request]]:
        pass



import hashlib
import os
import pickle
from typing import Any, Dict

from vidur.entities import Batch, Request
from vidur.scheduler.replica_scheduler.mnemosyne_fcfs_replica_scheduler import (
    MnemosyneFCFSReplicaScheduler,
    round_down_to_nearest_multiple,
    round_up_to_nearest_multiple,
)

PREFILL_TIME_PREDICTION_CACHE_GRANULARITY = 256
EXECUTION_TIME_PREDICTION_SLACK = 0.1
EXECUTION_TIME_PREDICTION_START_CHUNK_SIZE = 512
EXECUTION_TIME_PREDICTION_CHUNK_SIZE_GRANULARITY = 32
EXECUTION_TIME_PREDICTION_MAX_CHUNK_SIZE = 4096
EXECUTION_TIME_PREDICTION_MIN_CHUNK_SIZE = 32


class PrefillTimeCalculator:
    def __init__(
        self,
        execution_time_predictor,
        pipeline_parallel_size: int,
        kv_parallel_size: int,
        max_num_tokens_per_kvp_group: int,
        max_prefill_length: int,
        target_batch_time: float,
        enable_sequence_pipeline_parallelism: bool = False,
    ):
        self.execution_time_predictor = execution_time_predictor
        self.pipeline_parallel_size = pipeline_parallel_size
        self.kv_parallel_size = kv_parallel_size
        self.max_num_tokens_per_kvp_group = max_num_tokens_per_kvp_group
        self.target_batch_time = target_batch_time
        self.max_prefill_length = max_prefill_length
        self.enable_sequence_pipeline_parallelism = enable_sequence_pipeline_parallelism

        # Initialize cache for different prompt lengths
        cache_file_path = f"{execution_time_predictor._cache_dir}/prefill_time_cache_{self.get_hash()}.pkl"

        if os.path.exists(cache_file_path):
            with open(cache_file_path, "rb") as f:
                self.prefill_time_cache = pickle.load(f)
        else:
            self.prefill_time_cache = self._initialize_prefill_time_cache()
            with open(cache_file_path, "wb") as f:
                pickle.dump(self.prefill_time_cache, f)

    def _get_optimal_chunk_size(
        self,
        remaining_tokens: int,
        num_kvp_groups: int,
        num_kv_tokens: int,
    ) -> tuple[int, float]:
        # Binary search for optimal chunk size
        high = min(
            2 * EXECUTION_TIME_PREDICTION_START_CHUNK_SIZE,
            remaining_tokens,
            EXECUTION_TIME_PREDICTION_MAX_CHUNK_SIZE,
        )
        low = 0
        closest_match = 0
        closest_time = None
        seen_chunk_sizes = set()

        while low <= high:
            mid = (low + high) // 2

            if mid < remaining_tokens:
                mid = round_down_to_nearest_multiple(
                    mid, EXECUTION_TIME_PREDICTION_CHUNK_SIZE_GRANULARITY
                )
                if mid == 0:
                    mid = min(
                        remaining_tokens, EXECUTION_TIME_PREDICTION_MIN_CHUNK_SIZE
                    )
            else:
                mid = remaining_tokens

            if mid in seen_chunk_sizes or mid == 0:
                break

            seen_chunk_sizes.add(mid)

            execution_time = (
                self.execution_time_predictor.get_execution_time(
                    Batch(
                        0,  # replica_id
                        [None],  # requests
                        [mid],  # num_q_tokens
                        [num_kv_tokens],  # num_kv_tokens
                        [num_kvp_groups],  # num_active_kvp_groups
                        [1],  # last_kvp_group_ids
                        0,  # kvp_group_id
                    ),
                    pipeline_stage=0,
                ).total_time
                * self.pipeline_parallel_size
            )

            # Check if execution time is within slack range
            if execution_time >= self.target_batch_time * (
                1 - EXECUTION_TIME_PREDICTION_SLACK
            ) and execution_time <= self.target_batch_time * (
                1 + EXECUTION_TIME_PREDICTION_SLACK
            ):
                closest_match = mid
                closest_time = execution_time
                break
            elif execution_time < self.target_batch_time * (
                1 - EXECUTION_TIME_PREDICTION_SLACK
            ):
                low = mid
            else:
                high = mid

            # Keep track of closest match
            if closest_time is None or abs(
                execution_time - self.target_batch_time
            ) < abs(closest_time - self.target_batch_time):
                closest_match = mid
                closest_time = execution_time

        return closest_match, closest_time

    def _calculate_total_prefill_time(
        self,
        num_prefill_tokens: int,
    ) -> float:

        total_time = 0
        current_tokens = 0

        # Process tokens in chunks until all prefill is complete
        while current_tokens < num_prefill_tokens:
            # Get active KV parallel groups for current state
            num_kvp_groups = current_tokens // self.max_num_tokens_per_kvp_group + 1
            if num_kvp_groups > 1:
                num_kv_tokens = self.max_num_tokens_per_kvp_group
            else:
                num_kv_tokens = current_tokens

            remaining = num_prefill_tokens - current_tokens

            # Find optimal chunk size for target batch time
            chunk_size, execution_time = self._get_optimal_chunk_size(
                remaining_tokens=remaining,
                num_kvp_groups=num_kvp_groups,
                num_kv_tokens=num_kv_tokens,
            )

            if chunk_size == 0:
                break

            total_time += execution_time
            current_tokens += chunk_size

        print(f"Prefill time for {num_prefill_tokens} tokens: {total_time}")

        return total_time

    def _initialize_prefill_time_cache(self) -> Dict[int, float]:
        """
        Precompute execution times for different prompt lengths from
        CACHE_GRANULARITY to MAX_PROMPT_LENGTH with CACHE_GRANULARITY step size.
        """
        cache = {}
        cache[0] = 0

        for num_tokens in range(
            PREFILL_TIME_PREDICTION_CACHE_GRANULARITY,
            self.max_prefill_length + 1,
            PREFILL_TIME_PREDICTION_CACHE_GRANULARITY,
        ):
            cache[num_tokens] = self._calculate_total_prefill_time(num_tokens)

        return cache

    def get_prefill_time(
        self,
        num_prefill_tokens: int,
        num_processed_tokens: int = 0,
    ) -> float:
        """
        Calculate prefill time by using the difference between cached times
        for total tokens and processed tokens.
        """
        remaining_tokens = num_prefill_tokens - num_processed_tokens
        assert remaining_tokens > 0, "No tokens to prefill"

        # Get nearest cached sizes
        total_cached_size = round_up_to_nearest_multiple(
            num_prefill_tokens, PREFILL_TIME_PREDICTION_CACHE_GRANULARITY
        )
        processed_cached_size = round_down_to_nearest_multiple(
            num_processed_tokens, PREFILL_TIME_PREDICTION_CACHE_GRANULARITY
        )

        # Get base time from cache difference
        total_time = (
            self.prefill_time_cache[total_cached_size]
            - self.prefill_time_cache[processed_cached_size]
        )

        if self.enable_sequence_pipeline_parallelism:
            total_time = total_time / self.pipeline_parallel_size
        else:
            total_time = total_time

        return total_time

    def get_hash(self) -> str:
        attributes = {
            "pipeline_parallel_size": self.pipeline_parallel_size,
            "kv_parallel_size": self.kv_parallel_size,
            "max_num_tokens_per_kvp_group": self.max_num_tokens_per_kvp_group,
            "max_prefill_length": self.max_prefill_length,
            "target_batch_time": self.target_batch_time,
            **self.execution_time_predictor.to_dict(),
        }
        return hashlib.md5(str(attributes).encode("utf-8")).hexdigest()[0:8]


class MnemosyneEDFReplicaScheduler(MnemosyneFCFSReplicaScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._prefill_time_calculator = PrefillTimeCalculator(
            self._execution_time_predictor,
            self._num_stages,
            self._kv_parallel_size,
            self._max_num_tokens_per_kvp_group,
            self._request_generator_config.max_tokens,
            self._config.target_batch_time,
            self.is_sequence_pipeline_parallel_enabled,
        )

    def add_request(self, request):
        if not hasattr(request, "deadline"):
            prefill_time = self._prefill_time_calculator.get_prefill_time(
                request.num_prefill_tokens
            )
            deadline_time = prefill_time * self._config.deadline_multiplier
            deadline_time = max(prefill_time, self._config.min_deadline)
            request.deadline = request.arrived_at + deadline_time
            request.prefill_time = prefill_time
            request.deadline_time = deadline_time
        return super().add_request(request)

    def _get_request_priority(self, request: Request) -> Any:
        return (request.deadline, request.id)



from typing import Any, List

from vidur.entities import Batch, Request
from vidur.execution_time_predictor.base_execution_time_predictor import (
    BaseExecutionTimePredictor,
)
from vidur.scheduler.replica_scheduler.mnemosyne_base_replica_scheduler import (
    BatchFormationTracker,
    MnemosyneBaseReplicaScheduler,
)

EXECUTION_TIME_PREDICTION_SLACK = 0.1
EXECUTION_TIME_PREDICTION_START_CHUNK_SIZE = 512
EXECUTION_TIME_PREDICTION_CHUNK_SIZE_GRANULARITY = 32
EXECUTION_TIME_PREDICTION_MAX_CHUNK_SIZE = 4096
EXECUTION_TIME_PREDICTION_MIN_CHUNK_SIZE = 32


def round_down_to_nearest_multiple(value: int, multiple: int) -> int:
    return (value // multiple) * multiple


def round_up_to_nearest_multiple(value: int, multiple: int) -> int:
    return ((value + multiple - 1) // multiple) * multiple


class BatchFormationTrackerWithRuntimePrediction(BatchFormationTracker):
    def __init__(
        self,
        replica_id: int,
        max_micro_batch_size: int,
        pipeline_parallel_size: int,
        kv_parallel_size: int,
        max_num_tokens_per_kvp_group: int,
        execution_time_predictor: BaseExecutionTimePredictor,
    ):
        super().__init__(
            replica_id,
            max_micro_batch_size,
            kv_parallel_size,
            max_num_tokens_per_kvp_group,
        )
        self.pipeline_parallel_size = pipeline_parallel_size
        self.execution_time_predictor = execution_time_predictor

        self.batch_execution_time_predictions: List[int] = [
            0 for _ in range(self.kv_parallel_size)
        ]

    def add_request(
        self,
        request: Request,
        num_q_tokens: int,
        num_processed_tokens: int,
        active_kvp_groups: List[int],
    ) -> None:
        super().add_request(
            request, num_q_tokens, num_processed_tokens, active_kvp_groups
        )

        if num_q_tokens == 1:
            # Do not update predictions for decode requests
            # We are assuming that the decode requests are all added
            # at the beginning and don't need q chunk sizes, so we can just
            # do updates once we start adding prefills
            return

        for kvp_group_id in range(self.kv_parallel_size):
            self.batch_execution_time_predictions[kvp_group_id] = (
                self._compute_batch_execution_time(kvp_group_id)
            )

    def _compute_batch_execution_time(
        self,
        kvp_group_id: int,
        extra_requests: List[Request] = [],
        extra_num_q_tokens: List[int] = [],
        extra_num_kv_tokens: List[int] = [],
        extra_num_active_kvp_groups: List[int] = [],
        extra_last_kvp_group_ids: List[int] = [],
    ) -> int:
        if len(self.batch_requests[kvp_group_id]) + len(extra_requests) == 0:
            return 0

        return (
            self.execution_time_predictor.get_execution_time(
                Batch(
                    self.replica_id,
                    self.batch_requests[kvp_group_id] + extra_requests,
                    self.batch_num_q_tokens[kvp_group_id] + extra_num_q_tokens,
                    self.batch_num_kv_tokens[kvp_group_id] + extra_num_kv_tokens,
                    self.batch_num_active_kvp_groups[kvp_group_id]
                    + extra_num_active_kvp_groups,
                    self.batch_last_kvp_group_ids[kvp_group_id]
                    + extra_last_kvp_group_ids,
                    kvp_group_id,
                ),
                pipeline_stage=0,
            ).total_time
            * self.pipeline_parallel_size
        )

    def get_batch_execution_time(self, kvp_group_id: int) -> int:
        return self.batch_execution_time_predictions[kvp_group_id]

    def get_batch_execution_time_for_kvp_groups(self, kvp_group_ids: List[int]) -> int:
        return [
            self.batch_execution_time_predictions[kvp_group_id]
            for kvp_group_id in kvp_group_ids
        ]

    def get_max_chunk_size_for_request(
        self,
        request: Request,
        num_processed_tokens: int,
        active_kvp_groups: List[int],
        target_batch_time: float,
    ) -> int:
        # identify the kvp group with the maximum execution time, and get the execution time and group id
        max_execution_time_group_id = active_kvp_groups[0]
        max_execution_time = 0

        for kvp_group_id in active_kvp_groups:
            execution_time = self.get_batch_execution_time(kvp_group_id)
            if execution_time > max_execution_time:
                max_execution_time = execution_time
                max_execution_time_group_id = kvp_group_id

        if max_execution_time > target_batch_time * (
            1 - EXECUTION_TIME_PREDICTION_SLACK
        ):
            return 0

        is_last_group = max_execution_time_group_id == active_kvp_groups[-1]

        num_kv_tokens = self._get_num_kv_tokens(
            num_processed_tokens, active_kvp_groups, is_last_group
        )
        num_kvp_groups = len(active_kvp_groups)
        last_kvp_group_id = active_kvp_groups[-1]
        remaining_tokens = max(request.num_prefill_tokens - num_processed_tokens, 0)

        # Get initial bounds for binary search
        if hasattr(request, "__last_chunk_size"):
            high = request.__last_chunk_size
        else:
            high = EXECUTION_TIME_PREDICTION_START_CHUNK_SIZE

        # Cap high by remaining tokens and the prediction limit
        high = round_down_to_nearest_multiple(
            2 * high, EXECUTION_TIME_PREDICTION_CHUNK_SIZE_GRANULARITY
        )
        high = min(remaining_tokens, high, EXECUTION_TIME_PREDICTION_MAX_CHUNK_SIZE)
        low = 0

        # Binary search with 32-token steps except for last chunk
        closest_match = 0
        closest_time = None

        seen_chunk_sizes = set()

        while low <= high:
            mid = (low + high) // 2

            if mid < remaining_tokens:
                mid = round_down_to_nearest_multiple(
                    mid, EXECUTION_TIME_PREDICTION_CHUNK_SIZE_GRANULARITY
                )
                if mid == 0:
                    mid = min(
                        remaining_tokens, EXECUTION_TIME_PREDICTION_MIN_CHUNK_SIZE
                    )
            else:
                mid = remaining_tokens
            mid = min(mid, EXECUTION_TIME_PREDICTION_MAX_CHUNK_SIZE)

            if mid in seen_chunk_sizes:
                break

            seen_chunk_sizes.add(mid)

            if mid == 0:
                break

            execution_time = self._compute_batch_execution_time(
                max_execution_time_group_id,
                extra_requests=[request],
                extra_num_q_tokens=[mid],
                extra_num_kv_tokens=[num_kv_tokens],
                extra_num_active_kvp_groups=[num_kvp_groups],
                extra_last_kvp_group_ids=[last_kvp_group_id],
            )

            # Check if execution time is within both bounds of slack range
            if execution_time >= target_batch_time * (
                1 - EXECUTION_TIME_PREDICTION_SLACK
            ) and execution_time <= target_batch_time * (
                1 + EXECUTION_TIME_PREDICTION_SLACK
            ):
                # Found a good size within slack range
                closest_match = mid
                closest_time = execution_time
                break
            elif execution_time < target_batch_time * (
                1 - EXECUTION_TIME_PREDICTION_SLACK
            ):
                low = mid
            else:
                high = mid

            if closest_time is None or abs(execution_time - target_batch_time) < abs(
                closest_time - target_batch_time
            ):
                closest_match = mid
                closest_time = execution_time

        if closest_match != 0:
            request.__last_chunk_size = closest_match

        return closest_match


class MnemosyneFCFSReplicaScheduler(MnemosyneBaseReplicaScheduler):

    def _get_batch_formation_tracker(self) -> BatchFormationTracker:
        return BatchFormationTrackerWithRuntimePrediction(
            self._replica_id,
            self._max_micro_batch_size,
            self._num_stages,
            self._kv_parallel_size,
            self._max_num_tokens_per_kvp_group,
            self._execution_time_predictor,
        )

    def _get_request_next_num_q_tokens(
        self,
        time: float,
        request: Request,
        batch_formation_tracker: BatchFormationTracker,
    ) -> int:
        assert not request.completed

        if self._is_prefill_complete(request):
            return 1

        active_kvp_groups = self._get_active_kvp_group_ids(request)

        next_num_tokens = batch_formation_tracker.get_max_chunk_size_for_request(
            request,
            self._get_num_processed_tokens(request),
            active_kvp_groups,
            self._config.target_batch_time,
        )

        next_num_tokens = max(0, next_num_tokens)

        return next_num_tokens

    def _get_request_priority(self, request: Request) -> Any:
        return (request.arrived_at, request.id)



from vidur.execution_time_predictor.linear_regression_execution_time_predictor import (
    LinearRegressionExecutionTimePredictor,
)
from vidur.execution_time_predictor.random_forrest_execution_time_predictor import (
    RandomForrestExecutionTimePredictor,
)
from vidur.types import ExecutionTimePredictorType
from vidur.utils.base_registry import BaseRegistry


class ExecutionTimePredictorRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> ExecutionTimePredictorType:
        return ExecutionTimePredictorType.from_str(key_str)


ExecutionTimePredictorRegistry.register(
    ExecutionTimePredictorType.RANDOM_FORREST, RandomForrestExecutionTimePredictor
)
ExecutionTimePredictorRegistry.register(
    ExecutionTimePredictorType.LINEAR_REGRESSION, LinearRegressionExecutionTimePredictor
)



from typing import Dict, List

from vidur.entities.base_entity import BaseEntity
from vidur.entities.batch_stage import BatchStage, check_scheduled
from vidur.entities.execution_time import ExecutionTime
from vidur.entities.request import Request
from vidur.logger import init_logger

logger = init_logger(__name__)


class KVParallelBatchStage(BaseEntity):
    def __init__(
        self,
        replica_id: int,
        batch_stage_mapping: Dict[
            int, BatchStage
        ],  # map from kvp_group_id to batch_stage
        pipeline_stage: int,
        execution_time: ExecutionTime,
    ) -> None:
        self._id = KVParallelBatchStage.generate_id()
        self._replica_id = replica_id
        self._batch_stage_mapping = batch_stage_mapping
        self._pipeline_stage = pipeline_stage
        self._execution_time = execution_time

        self._total_execution_time = self._execution_time.total_time
        self._model_execution_time = self._execution_time.model_time

        self._scheduled_at = None
        self._completed_at = None
        self._scheduled = False

        self._requests = list(
            set(sum([x.requests for x in self._batch_stage_mapping.values()], []))
        )

    @property
    def replica_id(self) -> int:
        return self._replica_id

    @property
    def requests(self) -> List[Request]:
        return self._requests

    @property
    def batch_stage_mapping(self) -> Dict[int, BatchStage]:
        return self._batch_stage_mapping

    @property
    @check_scheduled
    def scheduled_at(self) -> float:
        return self._scheduled_at

    @property
    @check_scheduled
    def completed_at(self) -> float:
        return self._completed_at

    @property
    def execution_time(self) -> float:
        return self._total_execution_time

    @property
    def model_execution_time(self) -> float:
        return self._model_execution_time

    @property
    def pipeline_stage(self) -> int:
        return self._pipeline_stage

    def on_schedule(
        self,
        time: float,
    ) -> None:
        self._scheduled_at = time
        self._scheduled = True

        for batch_stage in self._batch_stage_mapping.values():
            batch_stage.on_schedule(time)

    def on_stage_end(
        self,
        time: float,
    ) -> None:
        self._completed_at = time

        for batch_stage in self._batch_stage_mapping.values():
            batch_stage.on_stage_end(time)

    def to_chrome_trace_events(self, time: int) -> dict:
        return sum(
            [
                batch_stage.to_chrome_trace_events(time)
                for batch_stage in self._batch_stage_mapping.values()
            ],
            [],
        )



import hashlib
from dataclasses import dataclass
from itertools import product
from typing import List, Optional


@dataclass
class ModelConfig:
    name: str
    identifier: str
    exclude_tp_dims: List[int] = None

    def get_key(self):
        return self.name

    def to_config_dict(self):
        return {
            "replica_config_model_name": self.identifier,
        }

    def is_tensor_parallel_degree_valid(self, tp_degree: int):
        return self.exclude_tp_dims is None or tp_degree not in self.exclude_tp_dims


@dataclass
class TraceConfig:
    name: str
    trace_file: str
    max_seq_len: int
    num_requests: int
    start_qps: float

    def get_key(self):
        return f"{self.name}_tk{self.max_seq_len}_rq{self.num_requests}"

    def to_config_dict(self):
        return {
            "request_generator_config_type": "synthetic",
            "length_generator_config_type": "trace",
            "interval_generator_config_type": "poisson",
            "synthetic_request_generator_config_max_tokens": self.max_seq_len,
            "trace_request_length_generator_config_max_tokens": self.max_seq_len,
            "zipf_request_length_generator_config_max_tokens": self.max_seq_len,
            "uniform_request_length_generator_config_max_tokens": self.max_seq_len,
            "fixed_request_length_generator_config_max_tokens": self.max_seq_len,
            "trace_request_generator_config_max_tokens": self.max_seq_len,
            "trace_request_length_generator_config_trace_file": self.trace_file,
            "trace_request_length_generator_config_prefill_scale_factor": 1,
            "trace_request_length_generator_config_decode_scale_factor": 1,
            "synthetic_request_generator_config_num_requests": self.num_requests,
            "vllm_scheduler_config_max_tokens_in_batch": self.max_seq_len,
        }


@dataclass
class ClusterConfig:
    device: str
    num_gpus: int
    gpus_per_node: int

    def get_key(self):
        return self.device

    def to_config_dict(self):
        return {
            "replica_config_device": self.device,
        }


@dataclass
class SchedulerConfig:
    scheduler: str
    chunk_size: Optional[int] = None

    def get_key(self):
        scheduler = self.scheduler

        if self.chunk_size is not None:
            scheduler += f"_cs{self.chunk_size}"

        return scheduler

    def to_config_dict(self):
        if self.scheduler == "vllm":
            return {
                "replica_scheduler_config_type": "vllm",
            }

        assert self.scheduler == "sarathi"
        assert self.chunk_size is not None
        return {
            "replica_scheduler_config_type": "sarathi",
            "sarathi_scheduler_config_chunk_size": self.chunk_size,
        }


class JobConfig:
    def __init__(
        self,
        model_config: ModelConfig,
        trace_config: TraceConfig,
        cluster_config: ClusterConfig,
        scheduler_config: SchedulerConfig,
        num_tensor_parallel_workers: int,
        num_pipeline_stages: int,
        batch_size: int,
    ):
        self.model_config = model_config
        self.trace_config = trace_config
        self.cluster_config = cluster_config
        self.scheduler_config = scheduler_config
        self.num_tensor_parallel_workers = num_tensor_parallel_workers
        self.num_pipeline_stages = num_pipeline_stages
        self.num_workers = self.num_tensor_parallel_workers * self.num_pipeline_stages
        self.batch_size = batch_size * num_pipeline_stages
        self.num_replicas = self.cluster_config.num_gpus // self.num_workers

        self.start_qps = self.trace_config.start_qps

    def is_valid(self):
        return (
            self.num_replicas > 0
            and self.model_config.is_tensor_parallel_degree_valid(
                self.num_tensor_parallel_workers
            )
            and self.num_tensor_parallel_workers <= self.cluster_config.gpus_per_node
        )

    def get_key(self):
        return (
            f"{self.model_config.name}_{self.trace_config.get_key()}_{self.cluster_config.get_key()}_{self.scheduler_config.get_key()}"
            f"_tp{self.num_tensor_parallel_workers}_pp{self.num_pipeline_stages}_bsz{self.batch_size}"
        )

    def get_human_readable_name(self):
        return (
            f"Model: {self.model_config.name}, Trace: {self.trace_config.name}, Cluster: {self.cluster_config.device}, "
            f"Scheduler: {self.scheduler_config.scheduler}, TP: {self.num_tensor_parallel_workers}, "
            f"PP: {self.num_pipeline_stages}, BSZ: {self.batch_size}, CS: {self.scheduler_config.chunk_size}, Hash: {self.get_hash()}"
        )

    def get_hash(self):
        return hashlib.sha1(self.get_key().encode("utf-8")).hexdigest()[:8]

    def to_config_dict(self):
        return {
            **self.model_config.to_config_dict(),
            **self.trace_config.to_config_dict(),
            **self.cluster_config.to_config_dict(),
            **self.scheduler_config.to_config_dict(),
            "replica_config_tensor_parallel_size": self.num_tensor_parallel_workers,
            "replica_config_num_pipeline_stages": self.num_pipeline_stages,
            "vllm_scheduler_config_batch_size_cap": self.batch_size,
            "lightllm_scheduler_config_batch_size_cap": self.batch_size,
            "orca_scheduler_config_batch_size_cap": self.batch_size,
            "faster_transformer_scheduler_config_batch_size_cap": self.batch_size,
            "sarathi_scheduler_config_batch_size_cap": self.batch_size,
            "cluster_config_num_replicas": self.num_replicas,
        }

    @classmethod
    def generate_job_configs(cls, config: dict):
        job_configs = []
        for (
            model_config,
            trace_config,
            cluster_config,
            scheduler_config,
            tp_dimension,
            pp_dimension,
            batch_size,
        ) in product(
            config["models"],
            config["traces"],
            config["clusters"],
            config["schedulers"],
            config["tp_dimensions"],
            config["pp_dimensions"],
            config["batch_sizes"],
        ):
            job_config = cls(
                ModelConfig(**model_config),
                TraceConfig(**trace_config),
                ClusterConfig(**cluster_config),
                SchedulerConfig(**scheduler_config),
                tp_dimension,
                pp_dimension,
                batch_size,
            )
            if not job_config.is_valid():
                continue

            job_configs.append(job_config)

        return job_configs

    @classmethod
    def generate_unique_model_job_configs(cls, config: dict, num_requests: int = 32):
        job_configs = []

        trace_config = TraceConfig(**config["traces"][0])
        trace_config.num_requests = num_requests
        scheduler_config = SchedulerConfig(**config["schedulers"][0])
        batch_size = config["batch_sizes"][0]
        pp_dimensions = config["pp_dimensions"]

        for model_config, cluster_config, tp_dimension, pp_dimension in product(
            config["models"],
            config["clusters"],
            config["tp_dimensions"],
            pp_dimensions,
        ):
            job_config = cls(
                ModelConfig(**model_config),
                trace_config,
                ClusterConfig(**cluster_config),
                scheduler_config,
                tp_dimension,
                pp_dimension,
                batch_size,
            )
            if not job_config.is_valid():
                continue

            job_configs.append(job_config)

        return job_configs


@dataclass
class SimulationConfig:
    output_dir: str
    cache_dir: str
    qps: float
    time_limit: int
    job_config: JobConfig

    def to_config_dict(self):
        return {
            **self.job_config.to_config_dict(),
            "metrics_config_output_dir": self.get_run_dir(),
            "linear_regression_execution_time_predictor_config_cache_dir": self.cache_dir,
            "random_forrest_execution_time_predictor_config_cache_dir": self.cache_dir,
            "poisson_request_interval_generator_config_qps": self.qps,
            "gamma_request_interval_generator_config_qps": self.qps,
            "time_limit": self.time_limit * 60,  # to seconds
            "no-metrics_config_save_table_to_wandb": None,
            "no-metrics_config_store_plots": None,
            "no-metrics_config_store_operation_metrics": None,
            "no-metrics_config_store_token_completion_metrics": None,
            "no-metrics_config_enable_chrome_trace": None,
        }

    def to_args(self):
        args = []

        for key, value in self.to_config_dict().items():
            if value is not None:
                args.append(f"--{key} {value}")
            else:
                args.append(f"--{key}")

        return " ".join(args)

    def to_human_readable_name(self):
        return f"{self.job_config.get_human_readable_name()}, QPS: {self.qps}"

    def get_run_dir(self):
        return f"{self.output_dir}/runs/{self.job_config.get_hash()}/{self.qps}"



import numpy as np
import plotly.graph_objs as go
import streamlit as st

from vidur.config_optimizer.analyzer.bottleneck_analyzer import BottleneckAnalyzer
from vidur.config_optimizer.analyzer.constants import AXIS_COLS, AXIS_COLS_LONG_TO_SHORT
from vidur.config_optimizer.analyzer.dashboard.utils import (
    add_advanced_filters,
    add_small_divider,
    convert_config_row_to_comparison_point,
    plot_cdf,
)


def get_best_configs(
    df,
    metric_1: str,
    percentile_1: str,
    slo_1: float,
    metric_2: str,
    percentile_2: str,
    slo_2: float,
    pick_min_if_no_match: bool = True,
):
    metric_1_col = f"{metric_1}_{percentile_1}%"
    metric_2_col = f"{metric_2}_{percentile_2}%"

    slo_df = df[(df[metric_1_col] <= slo_1) & (df[metric_2_col] <= slo_2)]

    if len(slo_df) == 0 and pick_min_if_no_match:
        min_metric_1 = df[metric_1_col].min()
        slo_df = df[(df[metric_1_col] == min_metric_1) & (df[metric_2_col] < slo_2)]
        if len(slo_df) == 0:
            min_metric_2 = df[metric_2_col].min()
            slo_df = df[
                (df[metric_1_col] == min_metric_1) & (df[metric_2_col] == min_metric_2)
            ]

    # group by config key and get row with max capacity_per_dollar
    best_configs_df = (
        slo_df.sort_values("QPS per Dollar", ascending=False)
        .groupby(["Model", "Trace"])
        .first()
        .reset_index()
    )

    return best_configs_df


def plot_parallel_coordinates(best_configs_df):
    best_configs_df["trace"] = (
        best_configs_df["Model"]
        + "<br>"
        + best_configs_df["Trace"]
        + "<br>"
        + best_configs_df["capacity_per_dollar_str"]
        + " QPS/$"
    )
    best_configs_df = best_configs_df.sort_values(["Model", "cost"])
    best_configs_df["trace_id"] = (
        best_configs_df.groupby(["Model", "cost"]).ngroup() + 1
    )
    labels = {**AXIS_COLS_LONG_TO_SHORT, "trace": "Trace"}

    dimensions = []

    for label_col, label_name in labels.items():
        if label_name == "Trace":
            dimension = go.parcats.Dimension(
                values=best_configs_df[label_col],
                label=label_name,
                categoryorder="array",
                categoryarray=best_configs_df["trace"].to_list(),
            )
        else:
            dimension = go.parcats.Dimension(
                values=best_configs_df[label_col],
                label=label_name,
                categoryorder="category ascending",
            )
        dimensions.append(dimension)

    # Create parcats trace
    color = np.log(best_configs_df["trace_id"])

    fig = go.Figure(
        data=[
            go.Parcats(
                dimensions=dimensions,
                line={
                    "color": color,
                    "colorscale": "agsunset",
                },
                hoverinfo="skip",
                labelfont={
                    "size": 18,
                },
                tickfont={
                    "size": 16,
                },
                arrangement="freeform",
            )
        ]
    )
    # reduce the width of the plot
    fig.update_layout(width=1100, height=100 * len(best_configs_df))
    # remove padding from left and add to right
    fig.update_layout(margin=dict(l=0, r=50, t=30, b=20))

    st.plotly_chart(fig, use_container_width=True)


def render_config_panels(df, bottleneck_analyzer):
    cols = st.columns(len(df))

    for idx, config_row in df.iterrows():
        col = cols[idx]
        with col:
            container = st.container(border=True)
            with container:
                st.markdown(f"### {config_row['Model']} - {config_row['Trace']}")

                add_small_divider()

                st.markdown("##### Configuration Details")
                for axis_col, value in config_row[AXIS_COLS.values()].items():
                    st.write(f"**{axis_col}**: {value}")

                add_small_divider()

                st.markdown("##### Performance Metrics")

                st.write(f"**TTFT (P95)**: {config_row['TTFT_95%']:.2f} s")
                st.write(f"**TBT (P99)**: {config_row['TBT_99%']:.2f} ms")
                st.write(f"**QPS per Dollar**: {config_row['QPS per Dollar']:.2f}")
                st.write(f"**MFU**: {config_row['mfu_mean']:.2f}")
                st.write(f"**Memory Usage**: {config_row['memory_usage_mean']:.2f}%")
                st.write(f"**Busy Time**: {config_row['busy_time_percent_mean']:.2f}%")

                bottleneck_message = bottleneck_analyzer.analyze(config_row)
                st.write(f"**Bottleneck**: {bottleneck_message}")

                add_small_divider()

                # add a button to add this config to comparison
                add_to_comparison = st.button(
                    "Add to Comparison", key=f"add_to_comparison_{idx}"
                )

                if add_to_comparison:
                    st.session_state.comparison_points.append(
                        convert_config_row_to_comparison_point(config_row)
                    )

    add_small_divider()

    df["Model-Trace"] = df["Model"] + "-" + df["Trace"]

    st.markdown("### CDFs")
    plot_cdf(df, y="TTFT_cdf", y_name="TTFT", color="Model-Trace")
    plot_cdf(df, y="TBT_cdf", y_name="TBT", color="Model-Trace")
    plot_cdf(df, y="batch_size_cdf", y_name="Batch Size", color="Model-Trace")
    plot_cdf(
        df, y="batch_num_tokens_cdf", y_name="Batch Num Tokens", color="Model-Trace"
    )


def render_best_config_selection_page(df):
    # create a subpage for best config selection
    st.markdown("## Best Config Selection")
    st.markdown(
        "This tool helps you find and visualize variations in best configuration for a given model model under different workloads."
    )

    add_small_divider()
    st.markdown("### Input Parameters")

    models = df["Model"].unique()

    selected_model = st.selectbox("Select Model", models)

    col1, col2 = st.columns(2)

    with col1:
        ttft_slo_percentile = st.selectbox(
            "TTFT SLO Percentile:", [50, 75, 90, 95, 99], index=2
        )
        ttft_slo_value = st.slider(
            "TTFT SLO Value (s)", min_value=0.1, max_value=10.0, value=2.0
        )

    with col2:
        tbt_slo_percentile = st.selectbox(
            "TBT SLO Percentile:", [50, 75, 90, 95, 99], index=4
        )
        tbt_slo_value = st.slider(
            "TBT SLO Value (ms)", min_value=10, max_value=2000, value=200
        )

    df = df[(df["Model"] == selected_model)]

    filtered_df = add_advanced_filters(df)

    best_configs_df = get_best_configs(
        filtered_df,
        "TTFT",
        ttft_slo_percentile,
        ttft_slo_value,
        "TBT",
        tbt_slo_percentile,
        tbt_slo_value,
    )

    bottleneck_analyzer = BottleneckAnalyzer(
        ttft_slo_percentile,
        ttft_slo_value,
        tbt_slo_percentile,
        tbt_slo_value,
    )

    st.markdown("### Results")

    add_small_divider()

    if len(best_configs_df) > 0:
        plot_parallel_coordinates(best_configs_df)
        render_config_panels(best_configs_df, bottleneck_analyzer)
    else:
        st.markdown("No config satisfies the SLOs")



CONFIG_KEY = [
    "replica_num_pipeline_stages",
    "replica_num_tensor_parallel_workers",
    "replica_device",
    "cluster_num_replicas",
    "replica_scheduler_batch_size_cap",
    "replica_scheduler_provider",
    "sarathi_scheduler_chunk_size",
]

METRICS = {
    "request_e2e_time_normalized": "Normalized Request Latency",
    "ttft": "Time to First Token",
    "tbt": "Time Between Tokens",
}

AXIS_COLS = {
    "replica_num_pipeline_stages": "Pipeline Parallel Dim",
    "replica_num_tensor_parallel_workers": "Tensor Parallel Dim",
    "replica_scheduler_provider": "Scheduler",
    "sarathi_scheduler_chunk_size": "Sarathi Chunk Size",
    "replica_scheduler_batch_size_cap": "Batch Size",
    "replica_device": "SKU",
}

AXIS_COLS_SHORT = {
    "replica_num_pipeline_stages": "PP Dim",
    "replica_num_tensor_parallel_workers": "TP Dim",
    "replica_scheduler_provider": "Scheduler",
    "sarathi_scheduler_chunk_size": "Chunk Size",
    "replica_scheduler_batch_size_cap": "BS",
    "replica_device": "SKU",
}

AXIS_COLS_LONG_TO_SHORT = {
    "Pipeline Parallel Dim": "PP Dim",
    "Tensor Parallel Dim": "TP Dim",
    "Scheduler": "Scheduler",
    "Sarathi Chunk Size": "Chunk Size",
    "Batch Size": "BS",
    "SKU": "SKU",
}

PRETTY_NAMES = {
    "sarathi": "Sarathi-Serve",
    "orca+": "Orca+",
    "vllm": "vLLM",
    "h100": "H100",
    "a100": "A100",
    "a40": "A40",
}

# Prices extracted from https://docs.coreweave.com/welcome-to-coreweave/resource-based-pricing on 7th Feb, 2024
# Prices are in $/hr
GPU_COSTS = {
    "h100": 4.25,
    "a100": 2.21,
    "a40": 1.28,
}

CPU_MACHINE_COST = 3.36

# define colors
COLORS = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
]
TRANS_COLORS = [c + "95" for c in COLORS]



from vidur.entities.batch import Batch
from vidur.scheduler.replica_scheduler.base_replica_scheduler import (
    BaseReplicaScheduler,
)


class OrcaReplicaScheduler(BaseReplicaScheduler):
    def on_batch_end(self, batch: Batch) -> None:
        super().on_batch_end(batch)

        for request in batch.requests:
            if request.completed:
                self.free(request.id)
            else:
                self._preempted_requests.append(request)

    def _get_next_batch(self, time: float) -> Batch:
        requests = []
        num_q_tokens = []
        num_kv_tokens = []

        # all preempted_requests will have prefill completed
        while self._preempted_requests:
            if len(requests) == self._max_batch_size:
                break

            request = self._preempted_requests.pop(0)
            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)

        while self._request_queue:
            if len(requests) == self._max_batch_size:
                break

            if not self.can_allocate(self._max_blocks_per_sequence):
                break

            request = self._request_queue.pop(0)

            self.allocate(request.id, self._max_blocks_per_sequence)
            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)

        if not requests:
            return

        return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)



from typing import List

from vidur.events import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class ReplicaStageScheduleEvent(BaseEvent):
    def __init__(self, time: float, replica_id: int, stage_id: int):
        super().__init__(time, EventType.REPLICA_STAGE_SCHEDULE)

        self._replica_id = replica_id
        self._stage_id = stage_id

        self._batch = None
        self._batch_stage = None
        self._is_last_stage = None

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.batch_stage_end_event import BatchStageEndEvent

        stage_scheduler = scheduler._replica_schedulers[
            self._replica_id
        ]._replica_stage_schedulers[self._stage_id]

        self._batch, self._batch_stage, execution_time = stage_scheduler.on_schedule()

        if not (self._batch and self._batch_stage):
            return []

        self._batch_stage.on_schedule(self.time)
        metrics_store.on_replica_stage_schedule(
            self.time,
            self._replica_id,
            self._stage_id,
            self._batch_stage,
            execution_time,
        )

        self._is_last_stage = stage_scheduler.is_last_stage

        return [
            BatchStageEndEvent(
                self.time + self._batch_stage.execution_time,
                self._replica_id,
                self._stage_id,
                self._is_last_stage,
                self._batch,
                self._batch_stage,
            ),
        ]

    def to_dict(self):
        return {
            "time": self.time,
            "event_type": self.event_type,
            "replica_id": self._replica_id,
            "stage_id": self._stage_id,
            "batch_id": self._batch.id if self._batch else None,
            "batch_stage_id": self._batch_stage.id if self._batch_stage else None,
            "is_last_stage": self._is_last_stage,
        }
